import torch.distributed as dist


def get_world_size():
    if not dist.is_available():
        return 1
    if not dist.is_initialized():
        return 1

    return dist.get_world_size()


def reduce_mean(tensor):
    '''
    Reduce the tensor across all machines, the operation is in-place.
    :param tensor: tensor to reduce
    :return: reduced tensor
    '''
    if not dist.is_available():
        return tensor
    if not dist.is_initialized():
        return tensor

    world_size = get_world_size()
    dist.all_reduce(tensor, op=dist.ReduceOp.SUM)
    return tensor.div_(world_size)


import time
import math
import torch
import torch.autograd as autograd

from .distributed import reduce_mean

from .cgd_utils import zero_grad, general_conjugate_gradient, Hvp_vec


class ACGD(object):
    def __init__(self, max_params, min_params,
                 max_reducer=None, min_reducer=None,
                 lr_max=1e-3, lr_min=1e-3,
                 backward_mode=False,
                 eps=1e-5, beta=0.99,
                 tol=1e-12, atol=1e-20,
                 solve_x=False, collect_info=False):
        self.backwardMode = backward_mode
        self.max_reducer = max_reducer
        self.min_reducer = min_reducer
        self.max_params = list(max_params)
        self.min_params = list(min_params)
        self.state = {'lr_max': lr_max, 'lr_min': lr_min,
                      'eps': eps, 'solve_x': solve_x,
                      'tol': tol, 'atol': atol,
                      'beta': beta, 'step': 0,
                      'old_max': None, 'old_min': None,  # start point of CG
                      'sq_exp_avg_max': None, 'sq_exp_avg_min': None}  # save last update
        self.info = {'grad_x': None, 'grad_y': None,
                     'hvp_x': None, 'hvp_y': None,
                     'cg_x': None, 'cg_y': None,
                     'time': 0, 'iter_num': 0}
        self.collect_info = collect_info

    def zero_grad(self):
        zero_grad(self.max_params)
        zero_grad(self.min_params)

    def get_info(self):
        if self.info['grad_x'] is None:
            print('Warning! No update information stored. Set collect_info=True before call this method')
        return self.info

    def state_dict(self):
        return self.state

    def load_state_dict(self, state_dict):
        self.state.update(state_dict)
        print('Load state: {}'.format(state_dict))

    def set_lr(self, lr_max, lr_min):
        self.state.update({'lr_max': lr_max, 'lr_min': lr_min})
        print('Maximizing side learning rate: {:.4f}\n '
              'Minimizing side learning rate: {:.4f}'.format(lr_max, lr_min))

    def step(self, loss, trigger):
        '''
        Update model weights using ACGD
        :param loss: objective function
        :param trigger: scalar, dummy loss to trigger DDP gradient reduction w.r.t. minimizer model parameters
        :return:
        '''
        lr_max = self.state['lr_max']
        lr_min = self.state['lr_min']
        beta = self.state['beta']
        eps = self.state['eps']
        tol = self.state['tol']
        atol = self.state['atol']
        time_step = self.state['step'] + 1
        self.state['step'] = time_step
        should_rebuild = time_step < 2

        grad_x = autograd.grad(loss, self.max_params, create_graph=True)
        grad_x_vec = torch.cat([g.contiguous().view(-1) for g in grad_x])
        grad_y = autograd.grad(loss, self.min_params, create_graph=True)
        grad_y_vec = torch.cat([g.contiguous().view(-1) for g in grad_y])

        grad_x_vec_d = grad_x_vec.clone().detach()
        grad_y_vec_d = grad_y_vec.clone().detach()
        # reduce leaf nodes across all devices
        grad_x_vec_d = reduce_mean(grad_x_vec_d)
        grad_y_vec_d = reduce_mean(grad_y_vec_d)

        sq_avg_x = self.state['sq_exp_avg_max']
        sq_avg_y = self.state['sq_exp_avg_min']
        sq_avg_x = torch.zeros_like(grad_x_vec_d, requires_grad=False) if sq_avg_x is None else sq_avg_x
        sq_avg_y = torch.zeros_like(grad_y_vec_d, requires_grad=False) if sq_avg_y is None else sq_avg_y
        sq_avg_x.mul_(beta).addcmul_(grad_x_vec_d, grad_x_vec_d, value=1-beta)
        sq_avg_y.mul_(beta).addcmul_(grad_y_vec_d, grad_y_vec_d, value=1-beta)

        bias_correction = 1 - beta ** time_step
        lr_max = math.sqrt(bias_correction) * lr_max / sq_avg_x.sqrt().add(eps)
        lr_min = math.sqrt(bias_correction) * lr_min / sq_avg_y.sqrt().add(eps)

        scaled_grad_x = torch.mul(lr_max, grad_x_vec_d)
        scaled_grad_y = torch.mul(lr_min, grad_y_vec_d)

        hvp_x_vec = Hvp_vec(grad_y_vec, self.max_params, scaled_grad_y,
                            backward=self.backwardMode,
                            retain_graph=True,
                            trigger=trigger,
                            rebuild=should_rebuild)  # h_xy * d_y
        hvp_y_vec = Hvp_vec(grad_x_vec, self.min_params, scaled_grad_x,
                            backward=self.backwardMode,
                            retain_graph=True,
                            trigger=trigger,
                            rebuild=should_rebuild)  # h_yx * d_x
        p_x = torch.add(grad_x_vec_d, - hvp_x_vec)
        p_y = torch.add(grad_y_vec_d, hvp_y_vec)
        if self.collect_info:
            norm_px = torch.norm(hvp_x_vec, p=2).item()
            norm_py = torch.norm(hvp_y_vec, p=2).item()
            timer = time.time()

        if self.state['solve_x']:
            p_y.mul_(lr_min.sqrt())
            cg_y, iter_num = general_conjugate_gradient(grad_x=grad_y_vec,
                                                        grad_y=grad_x_vec,
                                                        x_params=self.min_params,
                                                        y_params=self.max_params,
                                                        x_reducer=self.min_reducer,
                                                        y_reducer=self.max_reducer,
                                                        rebuild=should_rebuild,
                                                        trigger=trigger,
                                                        b=p_y, x=self.state['old_min'],
                                                        tol=tol, atol=atol,
                                                        lr_x=lr_min, lr_y=lr_max,
                                                        backward=self.backwardMode)
            old_min = cg_y.detach_()
            min_update = cg_y.mul(- lr_min.sqrt())
            hcg = Hvp_vec(grad_y_vec, self.max_params, min_update,
                          self.backwardMode,
                          trigger=trigger, reducer=self.max_reducer,
                          rebuild=should_rebuild).detach_()
            hcg.add_(grad_x_vec_d)
            max_update = hcg.mul(lr_max)
            old_max = hcg.mul(lr_max.sqrt())
        else:
            p_x.mul_(lr_max.sqrt())
            cg_x, iter_num = general_conjugate_gradient(grad_x=grad_x_vec,
                                                        grad_y=grad_y_vec,
                                                        x_params=self.max_params,
                                                        y_params=self.min_params,
                                                        x_reducer=self.max_reducer,
                                                        y_reducer=self.min_reducer,
                                                        rebuild=should_rebuild,
                                                        trigger=trigger,
                                                        b=p_x, x=self.state['old_max'],
                                                        tol=tol, atol=atol,
                                                        lr_x=lr_max, lr_y=lr_min,
                                                        backward=self.backwardMode)
            old_max = cg_x.detach_()
            max_update = cg_x.mul(lr_max.sqrt())
            hcg = Hvp_vec(grad_x_vec, self.min_params, max_update,
                          self.backwardMode,
                          trigger=trigger, reducer=self.min_reducer,
                          rebuild=should_rebuild).detach_()
            hcg.add_(grad_y_vec_d)
            min_update = hcg.mul(- lr_min)
            old_min = hcg.mul(lr_min.sqrt())
        self.state.update({'old_max': old_max, 'old_min': old_min,
                           'sq_exp_avg_max': sq_avg_x, 'sq_exp_avg_min': sq_avg_y})

        if self.collect_info:
            timer = time.time() - timer
            self.info.update({'time': timer, 'iter_num': iter_num,
                              'hvp_x': norm_px, 'hvp_y': norm_py})

        index = 0
        for p in self.max_params:
            p.data.add_(max_update[index: index + p.numel()].reshape(p.shape))
            index += p.numel()
        assert index == max_update.numel(), 'Maximizer CG size mismatch'

        index = 0
        for p in self.min_params:
            p.data.add_(min_update[index: index + p.numel()].reshape(p.shape))
            index += p.numel()
        assert index == min_update.numel(), 'Minimizer CG size mismatch'

        if self.collect_info:
            norm_gx = torch.norm(grad_x_vec, p=2).item()
            norm_gy = torch.norm(grad_y_vec, p=2).item()
            norm_cgx = torch.norm(max_update, p=2).item()
            norm_cgy = torch.norm(min_update, p=2).item()
            self.info.update({'grad_x': norm_gx, 'grad_y': norm_gy,
                              'cg_x': norm_cgx, 'cg_y': norm_cgy})
        self.state['solve_x'] = False if self.state['solve_x'] else True


'''
Generalized ACGD that works for general sum game
'''
import time
import math
import torch
import torch.autograd as autograd

from .gmres import GMRES
from .cgd_utils import zero_grad, \
    MvProd, \
    vectorize_grad

from functools import partial


class GACGD(object):
    def __init__(self, x_params, y_params,
                 x_reducer=None, y_reducer=None,
                 lr_x=1e-3, lr_y=1e-3,
                 eps=1e-5, beta=0.99,
                 tol=1e-5, atol=1e-6,
                 max_iter=None,
                 collect_info=False):
        self.x_reducer = x_reducer
        self.y_reducer = y_reducer
        self.x_params = list(x_params)
        self.y_params = list(y_params)
        self.max_iter = max_iter
        self.state = {'lr_x': lr_x, 'lr_y': lr_y,
                      'eps': eps,
                      'tol': tol, 'atol': atol,
                      'beta': beta, 'step': 0,
                      'x0': None,
                      'sq_exp_avg_x': None, 'sq_exp_avg_y': None}
        self.info = {'grad_x': None, 'grad_y': None,
                     'time': 0, 'iter_num': 0}
        self.collect_info = collect_info

    def zero_grad(self):
        zero_grad(self.x_params)
        zero_grad(self.y_params)

    def get_info(self):
        if self.info['grad_x'] is None:
            print('Warning! No update information stored. Set collect_info=True before call this method')
        return self.info

    def state_dict(self):
        return self.state

    def load_state_dict(self, state_dict, rank=None):
        keys = ['old_x', 'old_y', 'sq_exp_avg_max', 'sq_exp_avg_min']
        if rank:
            for key in keys:
                if key in state_dict:
                    state_dict[key] = state_dict[key].to(rank)
        self.state.update(state_dict)
        print('Load state: {}'.format(state_dict))

    def set_lr(self, lr_x, lr_y):
        self.state.update({'lr_x': lr_x, 'lr_y': lr_y})
        print('Maximizing side learning rate: {:.4f}\n '
              'Minimizing side learning rate: {:.4f}'.format(lr_x, lr_y))

    def step(self, loss_x, loss_y, trigger=None):
        '''
        Update model weights
        :param loss_x: objective function of x parameters
        :param loss_y: objective function of y parameters
        :param d_trigger: dummy loss to trigger DDP gradient reduction w.r.t. minimizer model parameters
        :param g_trigger: dummy loss to trigger DDP gradient reduction w.r.t. maximizer model parameters
        :return:
        '''
        if trigger is None:
            trigger = torch.zeros(1, device=loss_x.device)
        lr_x = self.state['lr_x']
        lr_y = self.state['lr_y']
        beta = self.state['beta']
        eps = self.state['eps']
        tol = self.state['tol']
        atol = self.state['atol']
        time_step = self.state['step'] + 1
        self.state['step'] = time_step
        should_rebuild = time_step < 2

        autograd.backward(loss_x, retain_graph=True, inputs=self.x_params)
        grad_x = vectorize_grad(self.x_params)

        autograd.backward(loss_y, retain_graph=True, inputs=self.y_params)
        grad_y = vectorize_grad(self.y_params)

        sq_avg_x = self.state['sq_exp_avg_x']
        sq_avg_y = self.state['sq_exp_avg_y']
        sq_avg_x = torch.zeros_like(grad_x, requires_grad=False) if sq_avg_x is None else sq_avg_x
        sq_avg_y = torch.zeros_like(grad_y, requires_grad=False) if sq_avg_y is None else sq_avg_y
        sq_avg_x.mul_(beta).addcmul_(grad_x, grad_x, value=1 - beta)
        sq_avg_y.mul_(beta).addcmul_(grad_y, grad_y, value=1 - beta)

        bias_correction = 1 - beta ** time_step
        lr_x = math.sqrt(bias_correction) * lr_x / sq_avg_x.sqrt().add(eps)
        lr_y = math.sqrt(bias_correction) * lr_y / sq_avg_y.sqrt().add(eps)
        # print(f'==== lr_x ====: {lr_x}')
        # print(f'==== lr_y ====: {lr_y}')
        # print()
        #
        # lr_x = torch.ones_like(grad_x)
        # lr_y = torch.ones_like(grad_y)
        scaled_grad_x = lr_x * grad_x
        scaled_grad_y = lr_y * grad_y
        RHS = torch.cat([scaled_grad_x, scaled_grad_y])

        grad_fy = autograd.grad(loss_x, self.y_params, create_graph=True)
        grad_fy_vec = torch.cat([g.contiguous().view(-1) for g in grad_fy])
        grad_gx = autograd.grad(loss_y, self.x_params, create_graph=True)
        grad_gx_vec = torch.cat([g.contiguous().view(-1) for g in grad_gx])

        # prev_x0 = self.state['x0']
        prev_x0 = None
        Avp = partial(MvProd,
                      grad_fy=grad_fy_vec, grad_gx=grad_gx_vec,
                      x_params=self.x_params, y_params=self.y_params,
                      lr_x=lr_x, lr_y=lr_y,
                      trigger=trigger,
                      x_reducer=self.x_reducer, y_reducer=self.y_reducer,
                      rebuild=should_rebuild)
        soln = GMRES(Avp=Avp, b=RHS, x0=prev_x0,
                     max_iter=self.max_iter,
                     tol=tol, atol=atol)
        # soln = CuGMRES(grad_fy=grad_fy_vec, grad_gx=grad_gx_vec,
        #                x_params=self.x_params, y_params=self.y_params,
        #                b=RHS, lr_x=lr_x, lr_y=lr_y,
        #                x0=None, max_iter=self.max_iter,
        #                tol=tol, trigger=trigger,
        #                x_reducer=self.x_reducer,
        #                y_reducer=self.y_reducer,
        #                rebuild=should_rebuild)
        self.state['x0'] = soln
        self.state.update(
            {
                'sq_exp_avg_x': sq_avg_x,
                'sq_exp_avg_y': sq_avg_y
            }
        )
        index = 0
        for p in self.x_params:
            size = p.numel()
            p.data.add_(- soln[index:index + size].reshape(p.shape))
            index += size
        for p in self.y_params:
            size = p.numel()
            p.data.add_(- soln[index:index + size].reshape(p.shape))
            index += size



import torch
import torch.autograd as autograd
import warnings


def vectorize_grad(params):
    '''
    Extract .grad field of parameters and concatenate them into a vector.
    If grad is None, it is replace with zeros.
    :param params: list of parameters
    :return: vector
    '''
    grad_list = []
    for p in params:
        if p.grad is not None:
            grad_list.append(p.grad.contiguous().view(-1))
            del p.grad
        else:
            # replace None with zeros
            grad_list.append(torch.zeros_like(p).view(-1))
    return torch.cat(grad_list)


def conjugate_gradient(grad_x, grad_y,
                       x_params, y_params,
                       b, x=None, nsteps=None,
                       tol=1e-10, atol=1e-16,
                       lr_x=1.0, lr_y=1.0,
                       device=torch.device('cpu')):
    """
    :param grad_x:
    :param grad_y:
    :param x_params:
    :param y_params:
    :param b: vec
    :param nsteps: max number of steps
    :param residual_tol:
    :return: A ** -1 * b
    h_1 = D_yx * p
    h_2 = D_xy * D_yx * p
    A = I + lr_x * D_xy * lr_y * D_yx
    """
    if nsteps is None:
        nsteps = b.shape[0]
    if x is None:
        x = torch.zeros(b.shape[0], device=b.device)
        r = b.clone().detach()
    else:
        h1 = Hvp_vec(grad_vec=grad_x, params=y_params, vec=x, retain_graph=True).detach_().mul(lr_y)
        h2 = Hvp_vec(grad_vec=grad_y, params=x_params, vec=h1, retain_graph=True).detach_().mul(lr_x)
        Avx = x + h2
        r = b.clone().detach() - Avx
        nsteps -= 1

    p = r.clone().detach()
    rdotr = torch.dot(r, r)
    residual_tol = tol * torch.dot(b, b)
    if rdotr < residual_tol or rdotr < atol:
        return x, 1


    for i in range(nsteps):
        # To compute Avp
        h_1 = Hvp_vec(grad_vec=grad_x, params=y_params, vec=p, retain_graph=True).detach_().mul(lr_y)
        h_2 = Hvp_vec(grad_vec=grad_y, params=x_params, vec=h_1, retain_graph=True).detach_().mul(lr_x)
        Avp_ = p + h_2

        alpha = rdotr / torch.dot(p, Avp_)
        x.data.add_(alpha * p)

        r.data.add_(- alpha * Avp_)
        new_rdotr = torch.dot(r, r)
        beta = new_rdotr / rdotr
        p = r + beta * p
        rdotr = new_rdotr
        if rdotr < residual_tol or rdotr < atol:
            break
    if i > 99:
        warnings.warn('CG iter num: %d' % (i + 1))
    return x, i + 1


def Hvp_vec(grad_vec, params, vec,
            backward=False,
            retain_graph=False,
            trigger=None,
            reducer=None,
            rebuild=False):
    '''
    Parameters:
        - grad_vec: Tensor of which the Hessian vector product will be computed
        - params: list of params, w.r.t which the Hessian will be computed
        - vec: The "vector" in Hessian vector product
        - retain_graph: keep the computation graph
        - trigger: scalar that will be used to trigger gradient reduction. Only needed when using DDP.
        - reducer: DDP reducer of the corresponding parameter
        - rebuild: If True, check last backward is reduced and rebuild bucket
    return: Hessian vector product
    '''
    if torch.isnan(grad_vec).any():
        raise ValueError('Gradvec nan')
    if torch.isnan(vec).any():
        raise ValueError('vector nan')
        # zero padding for None
    if backward:
        zero_grad(params)
        if reducer is not None:
            if rebuild:
                reducer._rebuild_buckets()
            reducer.prepare_for_backward([])
        autograd.backward(grad_vec + 0.0 * trigger, grad_tensors=vec,
                          inputs=params,
                          retain_graph=retain_graph)
        hvp = vectorize_grad(params)
    else:
        grad_grad = autograd.grad(grad_vec, params, grad_outputs=vec,
                                  retain_graph=retain_graph, create_graph=True,
                                  allow_unused=True)
        grad_list = []
        for i, p in enumerate(params):
            if grad_grad[i] is None:
                grad_list.append(torch.zeros_like(p).view(-1))
            else:
                grad_list.append(grad_grad[i].contiguous().view(-1))
        hvp = torch.cat(grad_list)
    if torch.isnan(hvp).any():
        raise ValueError('hvp Nan')
    return hvp


def general_conjugate_gradient(grad_x, grad_y,
                               x_params, y_params,
                               trigger,
                               b,
                               lr_x, lr_y,
                               x_reducer=None,
                               y_reducer=None,
                               rebuild=False,
                               backward=False,
                               x=None, nsteps=None,
                               tol=1e-10, atol=1e-16):
    '''
    Conjugate gradient algorithm for adaptive competitive gradient descent
    :param grad_x: grad w.r.t. x_params
    :param grad_y: grad w.r.t. y_params
    :param x_params: list of x parameters
    :param y_params: list of y parameters
    :param trigger: scalar, dummy loss term to trigger DDP comm hook
    :param b: the vector b in the linear system Ax=b
    :param lr_x: learning rate vector for x parameters
    :param lr_y: learning rate vector for y parameters
    :param x_reducer: reducer manager for x DDP module
    :param y_reducer: reducer manager for y DDP module
    :param rebuild: boolean, if True rebuild parameters for DDP. Only need for the first iteration.
    :param backward: True or False. If True, use backward to accumulate gradient.
                    Required to be True if working with DDP.
    :param x: initial guess of the solution
    :param nsteps: the maximum step of the CG inner loop
    :param tol: relative w.r.t. |b|^2 tolerance of the residual
    :param atol: absolute tolerance of the residual
    :return: (I + sqrt(lr_x) * D_xy * lr_y * D_yx * sqrt(lr_x)) ** -1 * b
    '''
    lr_x = lr_x.sqrt()
    if nsteps is None:
        nsteps = b.shape[0]
    if x is None:
        x = torch.zeros(b.shape[0], device=b.device)
        r = b.clone()
    else:
        h1 = Hvp_vec(grad_vec=grad_x, params=y_params,
                     vec=lr_x * x, backward=backward,
                     retain_graph=True,
                     trigger=trigger, reducer=y_reducer,
                     rebuild=rebuild).mul_(lr_y)
        h2 = Hvp_vec(grad_vec=grad_y, params=x_params,
                     vec=h1, backward=backward,
                     retain_graph=True,
                     trigger=trigger, reducer=x_reducer,
                     rebuild=rebuild).mul_(lr_x)
        Avx = x + h2
        r = b.clone() - Avx
        nsteps -= 1

    if grad_x.shape != b.shape:
        raise RuntimeError('CG: hessian vector product shape mismatch')

    p = r.clone().detach()
    rdotr = torch.dot(r, r)
    residual_tol = tol * torch.dot(b, b)
    if rdotr < residual_tol or rdotr < atol:
        return x, 1
    for i in range(nsteps):
        h_1 = Hvp_vec(grad_vec=grad_x, params=y_params,
                      vec=lr_x * p, backward=backward,
                      retain_graph=True,
                      trigger=trigger, reducer=y_reducer,
                      rebuild=rebuild).mul_(lr_y)
        h_2 = Hvp_vec(grad_vec=grad_y, params=x_params,
                      vec=h_1, backward=backward,
                      retain_graph=True,
                      trigger=trigger, reducer=x_reducer,
                      rebuild=rebuild).mul_(lr_x)
        Avp_ = p + h_2

        alpha = rdotr / torch.dot(p, Avp_)
        x.data.add_(alpha * p)
        r.data.add_(- alpha * Avp_)
        new_rdotr = torch.dot(r, r)
        beta = new_rdotr / rdotr
        rdotr = new_rdotr
        if rdotr < residual_tol or rdotr < atol:
            break
        p = r + beta * p
    if i > 100:
        warnings.warn('CG iter num: %d' % (i + 1))
    return x, i + 1


def zero_grad(params):
    for p in params:
        if p.grad is not None:
            p.grad.detach()
            p.grad.zero_()


def MvProd(vec,  # vector
           grad_fy,
           grad_gx,
           x_params,
           y_params,
           lr_x, lr_y,
           trigger=None,
           x_reducer=None,
           y_reducer=None,
           rebuild=False):
    '''
    Compute matrix vector product:
    \begin{pmatrix}
        I_m                                               & \eta_x \frac{\partial^2f}{\partial x \partial y}f \\
        \eta_y \frac{\partial^2g}{\partial y \partial x}g & I_n
    \end{pmatrix}
    \begin{pmatrix}
        b1 \\
        b2
    \end{pmatrix}

    Return:
        p1, p2
        where p1 =
    '''
    len_x = lr_x.shape[0]
    len_y = lr_y.shape[0]
    v1 = vec[0: len_x]
    v2 = vec[len_x: len_x + len_y]
    h1 = Hvp_vec(grad_fy, x_params, v2,
                 retain_graph=True, trigger=trigger,
                 reducer=x_reducer,
                 rebuild=rebuild)
    p1 = v1 + lr_x * h1
    h2 = Hvp_vec(grad_gx, y_params, v1,
                 retain_graph=True, trigger=trigger,
                 reducer=y_reducer,
                 rebuild=rebuild)
    p2 = v2 + lr_y * h2
    return torch.cat([p1, p2])



import time

import torch
import torch.autograd as autograd

from .cgd_utils import conjugate_gradient, Hvp_vec, zero_grad


class BCGD(object):
    def __init__(self, max_params, min_params,
                 lr_max=1e-3, lr_min=1e-3,
                 update_max=None,
                 tol=1e-12, atol=1e-20,
                 momentum=0.0, device=torch.device('cpu'),
                 solve_x=False, collect_info=True):
        self.max_params = list(max_params)
        self.min_params = list(min_params)
        self.state = {'lr_max': lr_max, 'lr_min': lr_min,
                      'momentum': momentum, 'solve_x': solve_x,
                      'tol': tol, 'atol': atol,
                      'step': 0, 'old_max': None, 'old_min': None,  # start point of CG
                      'exp_avg_max': 0.0, 'exp_avg_min': 0.0}  # save last update
        self.info = {'grad_x': None, 'grad_y': None,
                     'hvp_x': None, 'hvp_y': None,
                     'cg_x': None, 'cg_y': None,
                     'time': 0, 'iter_num': 0}
        self.update_max = update_max
        self.device = device
        self.collect_info = collect_info

    def zero_grad(self):
        zero_grad(self.max_params)
        zero_grad(self.min_params)

    def get_info(self):
        if self.info['grad_x'] is None:
            print('Warning! No update information stored. Set collect_info=True before call this method')
        return self.info

    def state_dict(self):
        return self.state

    def load_state_dict(self, state_dict):
        self.state.update(state_dict)
        print('Load state: {}'.format(state_dict))

    def set_lr(self, lr_max, lr_min):
        self.state.update({'lr_max': lr_max, 'lr_min': lr_min})
        print('Maximizing side learning rate: {:.4f}\n '
              'Minimizing side learning rate: {:.4f}'.format(lr_max, lr_min))

    def step(self, loss):
        """
            update rules:
            p_x = grad_x - lr * h_xy-v_y
            p_y = grad_y + lr * h_yx-v_x

            A_x = I + lr ** 2 * h_xy * h_yx
            A_y = I + lr ** 2 * h_yx * h_xy

            cg_x = A_x ** -1 * p_x
            cg_y = A_y ** -1 * p_y

            x = x + lr * cg_x
            y = y - lr * cg_y

            delta x = lr * cg_x
            delta y = - lr * cg_y

            given delta y, solve for delta x:
            delta x = - lr * (grad_x + h_xy-delta_y)
            cg_x = -(grad_x - lr * h_xy-cg_y)

            given delta x, solve for delta y
            delta y = lr * (grad_y + h_yx-delta_x)
            cg_y = -(grad_y + lr * h_yx-cg_x)

            p_x = grad_x - lr * h_xy-v_y
            A_x = I + lr ** 2 * h_xy * h_yx
            cg_x = A_x ** -1 * p_x
            cg_y = -(grad_y + lr * h_yx-cg_x)

            p_y = grad_y + lr * h_yx-v_x
            A_y = I + lr ** 2 * h_yx * h_xy
            cg_y = A_y ** -1 * p_y
            cg_x = -(grad_x - lr * h_xy-cg_y)
        """
        lr_max = self.state['lr_max']
        lr_min = self.state['lr_min']
        tol = self.state['tol']
        atol = self.state['atol']
        time_step = self.state['step'] + 1
        self.state['step'] = time_step

        grad_x = autograd.grad(loss, self.max_params, create_graph=True, retain_graph=True)
        grad_x_vec = torch.cat([g.contiguous().view(-1) for g in grad_x])
        grad_y = autograd.grad(loss, self.min_params, create_graph=True, retain_graph=True)
        grad_y_vec = torch.cat([g.contiguous().view(-1) for g in grad_y])
        grad_x_vec_d = grad_x_vec.clone().detach()
        grad_y_vec_d = grad_y_vec.clone().detach()

        if self.update_max is None:
            hvp_x_vec = Hvp_vec(grad_y_vec, self.max_params, grad_y_vec_d,
                                retain_graph=True)  # h_xy * d_y
            hvp_y_vec = Hvp_vec(grad_x_vec, self.min_params, grad_x_vec_d,
                                retain_graph=True)  # h_yx * d_x

            p_x = torch.add(grad_x_vec_d, - lr_min * hvp_x_vec).detach_()
            p_y = torch.add(grad_y_vec_d, lr_max * hvp_y_vec).detach_()
            if self.collect_info:
                norm_px = torch.norm(hvp_x_vec, p=2).item()
                norm_py = torch.norm(hvp_y_vec, p=2).item()
                timer = time.time()

            if self.state['solve_x']:
                cg_y, iter_num = conjugate_gradient(grad_x=grad_y_vec,
                                                    grad_y=grad_x_vec,
                                                    x_params=self.min_params,
                                                    y_params=self.max_params,
                                                    b=p_y, x=self.state['old_min'],
                                                    lr_x=lr_max, lr_y=lr_min,
                                                    tol=tol, atol=atol,
                                                    device=self.device)
                hcg = Hvp_vec(grad_y_vec, self.max_params, cg_y.detach_()).detach_()
                cg_x = torch.add(grad_x_vec_d, - lr_min * hcg)
            else:
                cg_x, iter_num = conjugate_gradient(grad_x=grad_x_vec,
                                                    grad_y=grad_y_vec,
                                                    x_params=self.max_params,
                                                    y_params=self.min_params,
                                                    b=p_x, x=self.state['old_max'],
                                                    lr_x=lr_max, lr_y=lr_min,
                                                    tol=tol, atol=atol,
                                                    device=self.device)
                hcg = Hvp_vec(grad_x_vec, self.min_params, cg_x.detach_()).detach_()
                cg_y = torch.add(grad_y_vec_d, lr_max * hcg)
            self.state.update({'old_max': cg_x, 'old_min': cg_y})

            if self.collect_info:
                timer = time.time() - timer
                self.info.update({'time': timer, 'iter_num': iter_num,
                                  'hvp_x': norm_px, 'hvp_y': norm_py})

            momentum = self.state['momentum']
            exp_avg_max, exp_avg_min = self.state['exp_avg_max'], self.state['exp_avg_min']
            if momentum != 0:
                bias_correction = 1 - momentum ** time_step
                lr_max /= bias_correction
                lr_min /= bias_correction
                cg_x = exp_avg_max.mul(momentum) + cg_x.mul(1 - momentum)
                cg_y = exp_avg_min.mul(momentum) + cg_y.mul(1 - momentum)
            index = 0
            for p in self.max_params:
                p.data.add_(lr_max * cg_x[index: index + p.numel()].reshape(p.shape))
                index += p.numel()
            assert index == cg_x.numel(), 'Maximizer CG size mismatch'
            index = 0
            for p in self.min_params:
                p.data.add_(- lr_min * cg_y[index: index + p.numel()].reshape(p.shape))
                index += p.numel()
            assert index == cg_y.numel(), 'Minimizer CG size mismatch'

            if self.collect_info:
                norm_gx = torch.norm(grad_x_vec, p=2).item()
                norm_gy = torch.norm(grad_y_vec, p=2).item()
                norm_cgx = torch.norm(cg_x, p=2).item()
                norm_cgy = torch.norm(cg_y, p=2).item()
                self.info.update({'grad_x': norm_gx, 'grad_y': norm_gy,
                                  'cg_x': norm_cgx, 'cg_y': norm_cgy})
            self.state['solve_x'] = False if self.state['solve_x'] else True
        elif self.update_max:
            hvp_x_vec = Hvp_vec(grad_y_vec, self.max_params, grad_y_vec_d,
                                retain_graph=True)  # h_xy * d_y
            p_x = torch.add(grad_x_vec_d, - lr_min * hvp_x_vec).detach_()
            cg, iter_num = conjugate_gradient(grad_x=grad_x_vec, grad_y=grad_y_vec,
                                              x_params=self.max_params,
                                              y_params=self.min_params, b=p_x, x=self.state['old_max'],
                                              nsteps=p_x.shape[0],
                                              lr_x=lr_max, lr_y=lr_min,
                                              tol=tol, atol=atol,
                                              device=self.device)
            cg.detach_()
            index = 0
            for p in self.max_params:
                p.data.add_(lr_max * cg[index: index + p.numel()].reshape(p.shape))
                index += p.numel()
            assert index == cg.numel(), 'Maximizer CG size mismatch'
            self.state.update({'old_max': cg})
        elif not self.update_max:
            hvp_y_vec = Hvp_vec(grad_x_vec, self.min_params, grad_x_vec_d,
                                retain_graph=True)  # h_yx * d_x
            p_y = torch.add(grad_y_vec_d, lr_max * hvp_y_vec).detach_()
            cg, iter_num = conjugate_gradient(grad_x=grad_y_vec, grad_y=grad_x_vec,
                                              x_params=self.min_params,
                                              y_params=self.max_params, b=p_y, x=self.state['old_min'],
                                              nsteps=p_y.shape[0],
                                              lr_x=lr_max, lr_y=lr_min,
                                              tol=tol, atol=atol,
                                              device=self.device)
            cg.detach_()
            index = 0
            for p in self.min_params:
                p.data.add_(- lr_min * cg[index: index + p.numel()].reshape(p.shape))
                index += p.numel()
            assert index == cg.numel(), 'Minimizer CG size mismatch'
            self.state.update({'old_min': cg})
        else:
            raise ValueError('Keyword argument update_max has to be None, True, or False')




import torch

def _check_nan(vec, msg):
    if torch.isnan(vec).any():
        raise ValueError(msg)


def _safe_normalize(x, threshold=None):
    norm = torch.norm(x)
    if threshold is None:
        threshold = torch.finfo(norm.dtype).eps
    normalized_x = x / norm if norm > threshold else torch.zeros_like(x)
    return normalized_x, norm


def arnoldi(vec,  # Matrix vector product
            V,  # List of existing basis
            H,  # H matrix
            j):  # number of basis
    '''
    Arnoldi iteration to find the j th l2-orthonormal vector
    compute the j-1 th column of Hessenberg matrix
    '''
    _check_nan(vec, 'Matrix vector product is Nan')

    for i in range(j):
        H[i, j - 1] = torch.dot(vec, V[i])
        vec = vec - H[i, j-1] * V[i]
    new_v, vnorm = _safe_normalize(vec)
    H[j, j - 1] = vnorm
    return new_v


def cal_rotation(a, b):
    '''
    Args:
        a: element h in position j
        b: element h in position j+1

    Returns:
        cosine = a / \sqrt{a^2 + b^2}
        sine = - b / \sqrt{a^2 + b^2}
    '''
    c = torch.sqrt(a * a + b * b)
    return a / c, - b / c


def apply_given_rotation(H, cs, ss, j):
    '''
    Apply givens rotation to H column
    :param H:
    :param cs:
    :param ss:
    :param j:
    :return:
    '''
    # apply previous rotation to the 0->j-1 columns
    for i in range(j):
        tmp = cs[i] * H[i, j] - ss[i] * H[i + 1, j]
        H[i + 1, j] = cs[i] * H[i+1, j] + ss[i] * H[i, j]
        H[i, j] = tmp
    cs[j], ss[j] = cal_rotation(H[j, j], H[j + 1, j])
    H[j, j] = cs[j] * H[j, j] - ss[j] * H[j + 1, j]
    H[j + 1, j] = 0
    return H, cs, ss


def GMRES(Avp,      # Linear operator
          b,        # tuple of two vectors, RHS of the linear system in which the first half has the same shape as grad_gx, the second half has the same shape as grad_fy
          x0=None,  # initial guess, tuple has the same shape as b
          max_iter=None,
          tol=1e-6,
          atol=1e-6):
    '''
    GMRES solver to solve
    $$
    \begin{pmatrix}
        \Delta x \\
        \Delta y
    \end{pmatrix} =
    \begin{pmatrix}
        I_m                                               & \eta_x \frac{\partial^2f}{\partial x \partial y}f \\
        \eta_y \frac{\partial^2g}{\partial y \partial x}g & I_n
    \end{pmatrix}^{-1}
    \begin{pmatrix}
        b1 \\
        b2
    \end{pmatrix}
    $$
    '''

    bnorm = torch.norm(b)
    _check_nan(b, 'RHS of the system is Nan')
    # if no inverse, it degrades to Adam with zero momentum
    if max_iter == 0 or bnorm < 1e-8:
        return b

    if max_iter is None:
        max_iter = b.shape[0]

    if x0 is None:
        x0 = torch.zeros_like(b)
        r0 = b
    else:
        r0 = b - Avp(x0)

    new_v, rnorm = _safe_normalize(r0)
    # initial guess residual
    beta = torch.zeros(max_iter + 1, device=b.device)
    beta[0] = rnorm

    V = []
    V.append(new_v)
    H = torch.zeros((max_iter + 1, max_iter + 1), device=b.device)
    cs = torch.zeros(max_iter, device=b.device)  # cosine values at each step
    ss = torch.zeros(max_iter, device=b.device)  # sine values at each step

    for j in range(max_iter):
        p = Avp(V[j])
        new_v = arnoldi(p, V, H, j + 1)  # Arnoldi iteration to get the j+1 th ba
        # sis
        _check_nan(new_v, f'{j}-th orthonormal basis contains Nan')
        V.append(new_v)

        H, cs, ss = apply_given_rotation(H, cs, ss, j)
        _check_nan(cs, f'{j}-th cosine contains NaN')
        _check_nan(ss, f'{j}-th sine contains NaN')
        beta[j + 1] = ss[j] * beta[j]
        beta[j] = cs[j] * beta[j]
        residual = torch.abs(beta[j + 1])
        if residual < tol * bnorm or residual < atol:
            print(f'\nGMRES iterations: {j}')
            break
        # print(f'{j}the gmres iteration:  residual {residual/bnorm}')
        if j == max_iter - 1:
            print(f'\nMax number of iterations: {max_iter}')
    y, _ = torch.triangular_solve(beta[0:j + 1].unsqueeze(-1), H[0:j + 1, 0:j + 1])  # j x j
    V = torch.stack(V[:-1], dim=0)
    sol = x0 + V.T @ y.squeeze(-1)
    return sol
